@@ -2,8 +2,9 @@
 #
 # Picard, the next-generation MusicBrainz tagger
 #
-# Copyright (C) 2014-2016, 2018-2019 Laurent Monin
+# Copyright (C) 2014-2016, 2018-2021 Laurent Monin
 # Copyright (C) 2017 Sambhav Kothari
+# Copyright (C) 2020 Philipp Wolfer
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -40,8 +41,8 @@ class CoverArtProviderCaaReleaseGroup(CoverArtProviderCaa):
     """Use cover art from album release group"""
 
     NAME = "CaaReleaseGroup"
-    TITLE = N_("CAA Release Group")
-    # FIXME: caa release group uses the same options than caa
+    TITLE = N_("Cover Art Archive: Release Group")
+    # FIXME: caa release group uses the same options than caa
     OPTIONS = None
 
     ignore_json_not_found_error = True
