@@ -6,7 +6,7 @@
 # Copyright (C) 2006-2009, 2011-2014, 2017 Lukáš Lalinský
 # Copyright (C) 2008 Gary van der Merwe
 # Copyright (C) 2008 amckinle
-# Copyright (C) 2008-2010, 2014-2015, 2018-2022 Philipp Wolfer
+# Copyright (C) 2008-2010, 2014-2015, 2018-2023 Philipp Wolfer
 # Copyright (C) 2009 Carlin Mangar
 # Copyright (C) 2010 Andrew Barnert
 # Copyright (C) 2011-2014 Michael Wiencek
@@ -22,11 +22,13 @@
 # Copyright (C) 2016 Suhas
 # Copyright (C) 2016-2018 Sambhav Kothari
 # Copyright (C) 2017-2018 Vishal Choudhary
-# Copyright (C) 2018 Bob Swift
 # Copyright (C) 2018 virusMac
+# Copyright (C) 2018, 2022-2023 Bob Swift
 # Copyright (C) 2019 Joel Lintunen
 # Copyright (C) 2020 Julius Michaelis
 # Copyright (C) 2020-2021 Gabriel Ferreira
+# Copyright (C) 2022 Kamil
+# Copyright (C) 2022 skelly37
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -45,13 +47,18 @@
 
 import argparse
 from functools import partial
+from hashlib import md5
 import logging
-import os.path
+import os
 import platform
 import re
 import shutil
 import signal
 import sys
+from textwrap import fill
+import time
+from urllib.parse import urlparse
+from uuid import uuid4
 
 from PyQt5 import (
     QtCore,
@@ -100,6 +107,7 @@ from picard.const.sys import (
 from picard.dataobj import DataObject
 from picard.disc import (
     Disc,
+    dbpoweramplog,
     eaclog,
     whipperlog,
 )
@@ -114,20 +122,29 @@ from picard.track import (
 )
 from picard.util import (
     check_io_encoding,
-    decode_filename,
     encode_filename,
     is_hidden,
     iter_files_from_objects,
     mbid_validate,
     normpath,
     periodictouch,
+    pipe,
     process_events_iter,
     system_supports_long_paths,
     thread,
     versions,
     webbrowser2,
 )
+from picard.util.cdrom import (
+    DISCID_NOT_LOADED_MESSAGE,
+    discid as _discid,
+    get_cdrom_drives,
+)
 from picard.util.checkupdate import UpdateCheckManager
+from picard.util.remotecommands import (
+    REMOTE_COMMANDS,
+    RemoteCommands,
+)
 from picard.webservice import WebService
 from picard.webservice.api_helpers import (
     AcoustIdAPIHelper,
@@ -170,6 +187,44 @@ def plugin_dirs():
     yield USER_PLUGIN_DIR
 
 
+class ParseItemsToLoad:
+
+    WINDOWS_DRIVE_TEST = re.compile(r"^[a-z]\:", re.IGNORECASE)
+
+    def __init__(self, items):
+        self.files = set()
+        self.mbids = set()
+        self.urls = set()
+
+        for item in items:
+            parsed = urlparse(item)
+            log.debug(f"Parsed: {repr(parsed)}")
+            if not parsed.scheme:
+                self.files.add(item)
+            if parsed.scheme == "file":
+                # remove file:// prefix safely
+                self.files.add(item[7:])
+            elif parsed.scheme == "mbid":
+                self.mbids.add(parsed.netloc + parsed.path)
+            elif parsed.scheme in {"http", "https"}:
+                # .path returns / before actual link
+                self.urls.add(parsed.path[1:])
+            elif IS_WIN and self.WINDOWS_DRIVE_TEST.match(item):
+                # Treat all single-character schemes as part of the file spec to allow
+                # specifying a drive identifier on Windows systems.
+                self.files.add(item)
+
+    # needed to indicate whether Picard should be brought to the front
+    def non_executable_items(self):
+        return bool(self.files or self.mbids or self.urls)
+
+    def __bool__(self):
+        return bool(self.files or self.mbids or self.urls)
+
+    def __str__(self):
+        return f"files: {repr(self.files)}  mbids: f{repr(self.mbids)}  urls: {repr(self.urls)}"
+
+
 class Tagger(QtWidgets.QApplication):
 
     tagger_stats_changed = QtCore.pyqtSignal()
@@ -184,7 +239,18 @@ class Tagger(QtWidgets.QApplication):
     _debug = False
     _no_restore = False
 
-    def __init__(self, picard_args, localedir, autoupdate):
+    def __init__(self, picard_args, localedir, autoupdate, pipe_handler=None):
+        # Initialize these variables early as they are needed for a clean
+        # shutdown.
+        self._acoustid = None
+        self.browser_integration = None
+        self.exit_cleanup = []
+        self.pipe_handler = None
+        self.priority_thread_pool = None
+        self.save_thread_pool = None
+        self.stopping = False
+        self.thread_pool = None
+        self.webservice = None
 
         super().__init__(sys.argv)
         self.__class__.__instance = self
@@ -192,7 +258,8 @@ class Tagger(QtWidgets.QApplication):
         config = get_config()
         theme.setup(self)
 
-        self._cmdline_files = picard_args.FILE
+        self._to_load = picard_args.processable
+
         self.autoupdate_enabled = autoupdate
         self._no_restore = picard_args.no_restore
         self._no_plugins = picard_args.no_plugins
@@ -206,6 +273,14 @@ class Tagger(QtWidgets.QApplication):
         # It's a valid reference, but its start() method doesn't work.
         self.thread_pool = QtCore.QThreadPool(self)
 
+        self.pipe_handler = pipe_handler
+
+        if self.pipe_handler:
+            self.pipe_handler.pipe_running = True
+            thread.run_task(self.pipe_server, self._pipe_server_finished)
+
+        self._init_remote_commands()
+
         # Provide a separate thread pool for operations that should not be
         # delayed by longer background processing tasks, e.g. because the user
         # expects instant feedback instead of waiting for a long list of
@@ -295,8 +370,6 @@ class Tagger(QtWidgets.QApplication):
         self.unclustered_files = UnclusteredFiles()
         self.nats = None
         self.window = MainWindow(disable_player=picard_args.no_player)
-        self.exit_cleanup = []
-        self.stopping = False
 
         # On macOS temporary files get deleted after 3 days not being accessed.
         # Touch these files regularly to keep them alive if Picard
@@ -308,10 +381,244 @@ class Tagger(QtWidgets.QApplication):
         if self.autoupdate_enabled:
             self.updatecheckmanager = UpdateCheckManager(parent=self.window)
 
+        thread.run_task(self.run_commands, self._run_commands_finished)
+
     @property
     def is_wayland(self):
         return self.platformName() == 'wayland'
 
+    def pipe_server(self):
+        IGNORED = {pipe.Pipe.MESSAGE_TO_IGNORE, pipe.Pipe.NO_RESPONSE_MESSAGE}
+        while self.pipe_handler.pipe_running:
+            messages = [x for x in self.pipe_handler.read_from_pipe() if x not in IGNORED]
+            if messages:
+                log.debug("pipe messages: %r", messages)
+                self.load_to_picard(messages)
+
+    def _pipe_server_finished(self, result=None, error=None):
+        if error:
+            log.error('pipe server failed: %r', error)
+        else:
+            log.debug('pipe server stopped')
+
+    def run_commands(self):
+        while not self.stopping:
+            if not RemoteCommands.command_queue.empty() and not RemoteCommands.get_running():
+                (cmd, arg) = RemoteCommands.command_queue.get()
+                if cmd in self.commands:
+                    arg = arg.strip()
+                    log.info("Executing command: %s %r", cmd, arg)
+                    if cmd == 'QUIT':
+                        thread.to_main(self.commands[cmd], arg)
+                    else:
+                        RemoteCommands.set_running(True)
+                        original_priority_thread_count = self.priority_thread_pool.activeThreadCount()
+                        original_main_thread_count = self.thread_pool.activeThreadCount()
+                        original_save_thread_count = self.save_thread_pool.activeThreadCount()
+                        thread.to_main_with_blocking(self.commands[cmd], arg)
+
+                        # Continue to show the task as running until all of the following
+                        # conditions are met:
+                        #
+                        #   - main thread pool active tasks count is less than or equal to the
+                        #     count at the start of task execution
+                        #
+                        #   - priority thread pool active tasks count is less than or equal to
+                        #     the count at the start of task execution
+                        #
+                        #   - save thread pool active tasks count is less than or equal to the
+                        #     count at the start of task execution
+                        #
+                        #   - there are no pending webservice requests
+                        #
+                        #   - there are no acoustid fingerprinting tasks running
+
+                        while True:
+                            time.sleep(0.1)
+                            if self.priority_thread_pool.activeThreadCount() > original_priority_thread_count or \
+                                    self.thread_pool.activeThreadCount() > original_main_thread_count or \
+                                    self.save_thread_pool.activeThreadCount() > original_save_thread_count or \
+                                    self.webservice.num_pending_web_requests or \
+                                    self._acoustid._running:
+                                continue
+                            break
+
+                        log.info("Completed command: %s %r", cmd, arg)
+                        RemoteCommands.set_running(False)
+
+                else:
+                    log.error("Unknown command: %r", cmd)
+                RemoteCommands.command_queue.task_done()
+            time.sleep(.01)
+
+    def _run_commands_finished(self, result=None, error=None):
+        if error:
+            log.error('command executor failed: %r', error)
+        else:
+            log.debug('command executor stopped')
+
+    @staticmethod
+    def load_to_picard(items):
+        commands = []
+        for item in items:
+            parts = str(item).split(maxsplit=1)
+            commands.append((parts[0], parts[1:] or ['']))
+        RemoteCommands.parse_commands_to_queue(commands)
+
+    def iter_album_files(self):
+        for album in self.albums.values():
+            yield from album.iterfiles()
+
+    def iter_all_files(self):
+        yield from self.unclustered_files.files
+        yield from self.iter_album_files()
+        yield from self.clusters.iterfiles()
+
+    def _init_remote_commands(self):
+        self.commands = {name: getattr(self, remcmd.method_name) for name, remcmd in REMOTE_COMMANDS.items()}
+
+    def handle_command_clear_logs(self, argstring):
+        self.window.log_dialog.clear()
+        self.window.history_dialog.clear()
+
+    def handle_command_cluster(self, argstring):
+        self.cluster(self.unclustered_files.files)
+
+    def handle_command_fingerprint(self, argstring):
+        for album_name in self.albums:
+            self.analyze(self.albums[album_name].iterfiles())
+
+    def handle_command_from_file(self, argstring):
+        RemoteCommands.get_commands_from_file(argstring)
+
+    def handle_command_load(self, argstring):
+        parsed_items = ParseItemsToLoad([argstring])
+        log.debug(str(parsed_items))
+
+        if parsed_items.files:
+            self.add_paths(parsed_items.files)
+
+        if parsed_items.urls or parsed_items.mbids:
+            file_lookup = self.get_file_lookup()
+            for item in parsed_items.mbids | parsed_items.urls:
+                file_lookup.mbid_lookup(item)
+
+    def handle_command_lookup(self, argstring):
+        if argstring:
+            argstring = argstring.upper()
+        if not argstring or argstring == 'ALL':
+            self.autotag(self.clusters)
+            self.autotag(self.unclustered_files.files)
+        elif argstring == 'CLUSTERED':
+            self.autotag(self.clusters)
+        elif argstring == 'UNCLUSTERED':
+            self.autotag(self.unclustered_files.files)
+        else:
+            log.error("Invalid LOOKUP command argument: '%s'", argstring)
+
+    def handle_command_lookup_cd(self, argstring):
+        if not _discid:
+            log.error(DISCID_NOT_LOADED_MESSAGE)
+            return
+        disc = Disc()
+        devices = get_cdrom_drives()
+
+        if not argstring:
+            if devices:
+                device = devices[0]
+            else:
+                device = None
+        elif argstring in devices:
+            device = argstring
+        else:
+            thread.run_task(
+                partial(self._parse_disc_ripping_log, disc, argstring),
+                partial(self._lookup_disc, disc),
+                traceback=self._debug)
+            return
+
+        thread.run_task(
+            partial(disc.read, encode_filename(device)),
+            partial(self._lookup_disc, disc),
+            traceback=self._debug)
+
+    def handle_command_pause(self, argstring):
+        arg = argstring.strip()
+        if arg:
+            try:
+                delay = float(arg)
+                if delay < 0:
+                    raise ValueError
+                log.debug("Pausing command execution by %d seconds.", delay)
+                thread.run_task(partial(time.sleep, delay))
+            except ValueError:
+                log.error(f"Invalid command pause time specified: {repr(argstring)}")
+        else:
+            log.error("No command pause time specified.")
+
+    def handle_command_quit(self, argstring):
+        if argstring.upper() == 'FORCE' or self.window.show_quit_confirmation():
+            self.quit()
+        else:
+            log.info("QUIT command cancelled by the user.")
+            RemoteCommands.set_quit(False)  # Allow queueing more commands.
+            return
+
+    def handle_command_remove(self, argstring):
+        for file in self.iter_all_files():
+            if file.filename == argstring:
+                self.remove([file])
+                return
+
+    def handle_command_remove_all(self, argstring):
+        for file in self.iter_all_files():
+            self.remove([file])
+
+    def handle_command_remove_empty(self, argstring):
+        _albums = [a for a in self.albums.values()]
+        for album in _albums:
+            if not any(album.iterfiles()):
+                self.remove_album(album)
+
+        for cluster in self.clusters:
+            if not any(cluster.iterfiles()):
+                self.remove_cluster(cluster)
+
+    def handle_command_remove_saved(self, argstring):
+        for track in self.iter_album_files():
+            if track.state == File.NORMAL:
+                self.remove([track])
+
+    def handle_command_remove_unclustered(self, argstring):
+        self.remove(self.unclustered_files.files)
+
+    def handle_command_save_matched(self, argstring):
+        for album in self.albums.values():
+            for track in album.iter_correctly_matched_tracks():
+                track.files[0].save()
+
+    def handle_command_save_modified(self, argstring):
+        for track in self.iter_album_files():
+            if track.state == File.CHANGED:
+                track.save()
+
+    def handle_command_scan(self, argstring):
+        self.analyze(self.unclustered_files.files)
+
+    def handle_command_show(self, argstring):
+        self.bring_tagger_front()
+
+    def handle_command_submit_fingerprints(self, argstring):
+        self.acoustidmanager.submit()
+
+    def handle_command_write_logs(self, argstring):
+        try:
+            with open(argstring, 'w', encoding='utf8') as f:
+                for x in self.window.log_dialog.log_tail.contents():
+                    f.write(f"{x.message}\n")
+        except Exception as e:
+            log.error("Error writing logs to a file: %s", e)
+
     def enable_menu_icons(self, enabled):
         self.setAttribute(QtCore.Qt.ApplicationAttribute.AA_DontShowIconsInMenus, not enabled)
 
@@ -400,25 +707,36 @@ class Tagger(QtWidgets.QApplication):
         if nat.loaded:
             self.nats.update()
 
+    def quit(self):
+        self.exit()
+        super().quit()
+
     def exit(self):
         if self.stopping:
             return
         self.stopping = True
         log.debug("Picard stopping")
-        self._acoustid.done()
-        self.thread_pool.waitForDone()
-        self.save_thread_pool.waitForDone()
-        self.priority_thread_pool.waitForDone()
-        self.browser_integration.stop()
-        self.webservice.stop()
+        if self._acoustid:
+            self._acoustid.done()
+        if self.pipe_handler:
+            self.pipe_handler.stop()
+        if self.webservice:
+            self.webservice.stop()
+        if self.thread_pool:
+            self.thread_pool.waitForDone()
+        if self.save_thread_pool:
+            self.save_thread_pool.waitForDone()
+        if self.priority_thread_pool:
+            self.priority_thread_pool.waitForDone()
+        if self.browser_integration:
+            self.browser_integration.stop()
         self.run_cleanup()
         QtCore.QCoreApplication.processEvents()
 
     def _run_init(self):
-        if self._cmdline_files:
-            files = [decode_filename(f) for f in self._cmdline_files]
-            self.add_paths(files)
-            del self._cmdline_files
+        if self._to_load:
+            self.load_to_picard(self._to_load)
+        del self._to_load
 
     def run(self):
         self.update_browser_integration()
@@ -568,14 +886,14 @@ class Tagger(QtWidgets.QApplication):
         for filename in filenames:
             filename = normpath(filename)
             if ignore_hidden and is_hidden(filename):
-                log.debug("File ignored (hidden): %r" % (filename))
+                log.debug("File ignored (hidden): %r", filename)
                 continue
             # Ignore .smbdelete* files which Applie iOS SMB creates by renaming a file when it cannot delete it
             if os.path.basename(filename).startswith(".smbdelete"):
                 log.debug("File ignored (.smbdelete): %r", filename)
                 continue
             if ignoreregex is not None and ignoreregex.search(filename):
-                log.info("File ignored (matching %r): %r" % (pattern, filename))
+                log.info("File ignored (matching %r): %r", pattern, filename)
                 continue
             if filename not in self.files:
                 file = open_file(filename)
@@ -858,7 +1176,9 @@ class Tagger(QtWidgets.QApplication):
     def lookup_discid_from_logfile(self):
         file_chooser = QtWidgets.QFileDialog(self.window)
         file_chooser.setNameFilters([
-            _("EAC / XLD / Whipper log files") + " (*.log)",
+            _("All supported log files") + " (*.log, *.txt)",
+            _("EAC / XLD / Whipper / fre:ac log files") + " (*.log)",
+            _("dBpoweramp log files") + " (*.txt)",
             _("All files") + " (*)",
         ])
         if file_chooser.exec_():
@@ -871,16 +1191,23 @@ class Tagger(QtWidgets.QApplication):
                 traceback=self._debug)
 
     def _parse_disc_ripping_log(self, disc, path):
-        try:
-            log.debug('Trying to parse "%s" as EAC / XLD log...', path)
-            toc = eaclog.toc_from_file(path)
-        except Exception:
+        log_readers = (
+            eaclog.toc_from_file,
+            whipperlog.toc_from_file,
+            dbpoweramplog.toc_from_file,
+        )
+        for reader in log_readers:
+            module_name = reader.__module__
             try:
-                log.debug('Trying to parse "%s" as Whipper log...', path)
-                toc = whipperlog.toc_from_file(path)
+                log.debug('Trying to parse "%s" with %s...', path, module_name)
+                toc = reader(path)
+                break
             except Exception:
-                log.warning('Failed parsing ripping log "%s"', path, exc_info=True)
-                raise
+                log.debug('Failed parsing ripping log "%s" with %s', path, module_name, exc_info=True)
+        else:
+            msg = N_('Failed parsing ripping log "%s"')
+            log.warning(msg, path)
+            raise Exception(_(msg) % path)
         disc.put(toc)
 
     @property
@@ -940,16 +1267,15 @@ class Tagger(QtWidgets.QApplication):
             log.error('Error while clustering: %r', error)
             return
 
-        with self.window.ignore_selection_changes:
-            self.window.set_sorting(False)
-            for file_cluster in process_events_iter(result):
-                files = set(file_cluster.files)
-                if len(files) > 1:
-                    cluster = self.load_cluster(file_cluster.title, file_cluster.artist)
-                else:
-                    cluster = self.unclustered_files
-                cluster.add_files(files)
-            self.window.set_sorting(True)
+        self.window.set_sorting(False)
+        for file_cluster in process_events_iter(result):
+            files = set(file_cluster.files)
+            if len(files) > 1:
+                cluster = self.load_cluster(file_cluster.title, file_cluster.artist)
+            else:
+                cluster = self.unclustered_files
+            cluster.add_files(files)
+        self.window.set_sorting(True)
 
         if callback:
             callback()
@@ -999,22 +1325,41 @@ class Tagger(QtWidgets.QApplication):
 
     def sighandler(self):
         self.signalnotifier.setEnabled(False)
-        self.exit()
         self.quit()
         self.signalnotifier.setEnabled(True)
 
 
-def version():
-    print("%s %s %s" % (PICARD_ORG_NAME, PICARD_APP_NAME, PICARD_FANCY_VERSION_STR))
+def print_help_for_commands():
+    maxwidth = 80
+    helpcmd = []
+    for name in sorted(REMOTE_COMMANDS):
+        remcmd = REMOTE_COMMANDS[name]
+        s = "  - %-34s %s" % (name + " " + remcmd.help_args, remcmd.help_text)
+        helpcmd.append(fill(s, width=maxwidth, subsequent_indent=' '*39))
 
+    print("""usage: picard -e [command] [arguments ...]
+    or picard -e [command 1] [arguments ...] -e [command 2] [arguments ...]
 
-def longversion():
-    print(versions.as_string())
+List of the commands available to execute in Picard from the command-line:
+""")
+    print("\n".join(helpcmd))
+
+    def fmt(s):
+        print(fill(s, width=maxwidth, initial_indent=' '*2))
+
+    fmt("Commands are case insensitive.")
+    fmt("Picard will try to load all the positional arguments before processing commands.")
+    fmt("If there is no instance to pass the arguments to, Picard will start and process the commands after the "
+        "positional arguments are loaded, as mentioned above. Otherwise they will be handled by the running "
+        "Picard instance")
+    fmt("Arguments are optional, but some commands may require one or more arguments to actually do something.")
 
 
 def process_picard_args():
     parser = argparse.ArgumentParser(
-        epilog="If one of the filenames begins with a hyphen, use -- to separate the options from the filenames."
+        formatter_class=argparse.RawDescriptionHelpFormatter,
+        epilog="""If one of the filenames begins with a hyphen, use -- to separate the options from the filenames.
+If a new instance will not be spawned files/directories will be passed to the existing instance"""
     )
     # Qt default arguments. Parse them so Picard does not interpret the
     # arguments as file names to load.
@@ -1029,6 +1374,10 @@ def process_picard_args():
                         help="location of the configuration file")
     parser.add_argument("-d", "--debug", action='store_true',
                         help="enable debug-level logging")
+    parser.add_argument("-e", "--exec", nargs="+", action='append',
+                        help="send command (arguments can be entered after space) to a running instance "
+                        "(use `-e help` for a list of the available commands)",
+                        metavar="COMMAND")
     parser.add_argument("-M", "--no-player", action='store_true',
                         help="disable built-in media player")
     parser.add_argument("-N", "--no-restore", action='store_true',
@@ -1037,16 +1386,37 @@ def process_picard_args():
                         help="do not load any plugins")
     parser.add_argument("--no-crash-dialog", action='store_true',
                         help="disable the crash dialog")
+    parser.add_argument("-s", "--stand-alone-instance", action='store_true',
+                        help="force Picard to create a new, stand-alone instance")
     parser.add_argument('-v', '--version', action='store_true',
                         help="display version information and exit")
     parser.add_argument("-V", "--long-version", action='store_true',
                         help="display long version information and exit")
-    parser.add_argument('FILE', nargs='*')
+    parser.add_argument('FILE_OR_URL', nargs='*',
+                        help="the file(s), URL(s) and MBID(s) to load")
 
-    return parser.parse_known_args()[0]
+    args = parser.parse_args()
+    args.remote_commands_help = False
+
+    args.processable = []
+    for x in args.FILE_OR_URL:
+        if not urlparse(x).netloc:
+            x = os.path.abspath(x)
+        args.processable.append(f"LOAD {x}")
+
+    if args.exec:
+        for e in args.exec:
+            args.remote_commands_help = args.remote_commands_help or "HELP" in {x.upper().strip() for x in e}
+            remote_command_args = e[1:] or ['']
+            for arg in remote_command_args:
+                args.processable.append(f"{e[0]} {arg}")
+
+    return args
 
 
 def main(localedir=None, autoupdate=True):
+    EXIT_NO_NEW_INSTANCE = 30403
+
     # Some libs (ie. Phonon) require those to be set
     QtWidgets.QApplication.setApplicationName(PICARD_APP_NAME)
     QtWidgets.QApplication.setOrganizationName(PICARD_ORG_NAME)
@@ -1068,10 +1438,38 @@ def main(localedir=None, autoupdate=True):
     signal.signal(signal.SIGINT, signal.SIG_DFL)
 
     picard_args = process_picard_args()
-    if picard_args.version:
-        return version()
+
     if picard_args.long_version:
-        return longversion()
+        return print(versions.as_string())
+    if picard_args.version:
+        return print(f"{PICARD_ORG_NAME} {PICARD_APP_NAME} {PICARD_FANCY_VERSION_STR}")
+    if picard_args.remote_commands_help:
+        return print_help_for_commands()
+
+    # any of the flags that change Picard's workflow significantly should trigger creation of a new instance
+    if picard_args.stand_alone_instance:
+        identifier = uuid4().hex
+    else:
+        identifier = md5(picard_args.config_file.encode('utf8')).hexdigest() if picard_args.config_file else 'main'  # nosec: B303
+        identifier += '_NP' if picard_args.no_plugins else ''
+
+    if picard_args.processable:
+        log.info("Sending messages to main instance: %r", picard_args.processable)
+
+    try:
+        pipe_handler = pipe.Pipe(app_name=PICARD_APP_NAME, app_version=PICARD_FANCY_VERSION_STR,
+                                    identifier=identifier, args=picard_args.processable)
+        should_start = pipe_handler.is_pipe_owner
+    except pipe.PipeErrorNoPermission as err:
+        log.error(err)
+        pipe_handler = None
+        should_start = True
+
+    # pipe has sent its args to existing one, doesn't need to start
+    if not should_start:
+        log.debug("No need for spawning a new instance, exiting...")
+        # just a custom exit code to show that picard instance wasn't created
+        sys.exit(EXIT_NO_NEW_INSTANCE)
 
     try:
         from PyQt5.QtDBus import QDBusConnection
@@ -1080,7 +1478,7 @@ def main(localedir=None, autoupdate=True):
     except ImportError:
         pass
 
-    tagger = Tagger(picard_args, localedir, autoupdate)
+    tagger = Tagger(picard_args, localedir, autoupdate, pipe_handler=pipe_handler)
 
     # Initialize Qt default translations
     translator = QtCore.QTranslator()
@@ -1093,4 +1491,9 @@ def main(localedir=None, autoupdate=True):
         log.debug('Qt locale %s not available', locale.name())
 
     tagger.startTimer(1000)
-    sys.exit(tagger.run())
+    exit_code = tagger.run()
+
+    if tagger.pipe_handler.unexpected_removal:
+        os._exit(exit_code)
+
+    sys.exit(exit_code)
