@@ -3,10 +3,10 @@
 # Picard, the next-generation MusicBrainz tagger
 #
 # Copyright (C) 2013 Michael Wiencek
-# Copyright (C) 2014-2015, 2018, 2020-2021 Laurent Monin
+# Copyright (C) 2014-2015, 2018, 2020-2022 Laurent Monin
 # Copyright (C) 2016-2017 Sambhav Kothari
-# Copyright (C) 2018 Philipp Wolfer
 # Copyright (C) 2018 Vishal Choudhary
+# Copyright (C) 2018, 2022 Philipp Wolfer
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -23,8 +23,6 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
 
-import locale
-
 from PyQt5 import (
     QtCore,
     QtGui,
@@ -35,6 +33,7 @@ from picard.collection import (
     load_user_collections,
     user_collections,
 )
+from picard.util import strxfrm
 
 
 class CollectionMenu(QtWidgets.QMenu):
@@ -51,7 +50,7 @@ class CollectionMenu(QtWidgets.QMenu):
         self.actions = []
         for id_, collection in sorted(user_collections.items(),
                                       key=lambda k_v:
-                                      (locale.strxfrm(str(k_v[1])), k_v[0])):
+                                      (strxfrm(str(k_v[1])), k_v[0])):
             action = QtWidgets.QWidgetAction(self)
             action.setDefaultWidget(CollectionMenuItem(self, collection))
             self.addAction(action)
@@ -102,10 +101,10 @@ class CollectionMenuItem(QtWidgets.QWidget):
         layout = QtWidgets.QVBoxLayout(self)
         style = self.style()
         layout.setContentsMargins(
-            style.pixelMetric(QtWidgets.QStyle.PM_LayoutLeftMargin),
-            style.pixelMetric(QtWidgets.QStyle.PM_FocusFrameVMargin),
-            style.pixelMetric(QtWidgets.QStyle.PM_LayoutRightMargin),
-            style.pixelMetric(QtWidgets.QStyle.PM_FocusFrameVMargin))
+            style.pixelMetric(QtWidgets.QStyle.PixelMetric.PM_LayoutLeftMargin),
+            style.pixelMetric(QtWidgets.QStyle.PixelMetric.PM_FocusFrameVMargin),
+            style.pixelMetric(QtWidgets.QStyle.PixelMetric.PM_LayoutRightMargin),
+            style.pixelMetric(QtWidgets.QStyle.PixelMetric.PM_FocusFrameVMargin))
         self.checkbox = CollectionCheckBox(self, menu, collection)
         layout.addWidget(self.checkbox)
 
@@ -118,7 +117,7 @@ class CollectionMenuItem(QtWidgets.QWidget):
         self.active = active
         palette = self.palette()
         textcolor = self.highlight_color if active else self.text_color
-        palette.setColor(QtGui.QPalette.WindowText, textcolor)
+        palette.setColor(QtGui.QPalette.ColorRole.WindowText, textcolor)
         self.checkbox.setPalette(palette)
 
     def enterEvent(self, e):
@@ -131,12 +130,12 @@ class CollectionMenuItem(QtWidgets.QWidget):
         painter = QtWidgets.QStylePainter(self)
         option = QtWidgets.QStyleOptionMenuItem()
         option.initFrom(self)
-        option.state = QtWidgets.QStyle.State_None
+        option.state = QtWidgets.QStyle.StateFlag.State_None
         if self.isEnabled():
-            option.state |= QtWidgets.QStyle.State_Enabled
+            option.state |= QtWidgets.QStyle.StateFlag.State_Enabled
         if self.active:
-            option.state |= QtWidgets.QStyle.State_Selected
-        painter.drawControl(QtWidgets.QStyle.CE_MenuItem, option)
+            option.state |= QtWidgets.QStyle.StateFlag.State_Selected
+        painter.drawControl(QtWidgets.QStyle.ControlElement.CE_MenuItem, option)
 
 
 class CollectionCheckBox(QtWidgets.QCheckBox):
@@ -148,11 +147,11 @@ class CollectionCheckBox(QtWidgets.QCheckBox):
 
         releases = collection.releases & menu.ids
         if len(releases) == len(menu.ids):
-            self.setCheckState(QtCore.Qt.Checked)
+            self.setCheckState(QtCore.Qt.CheckState.Checked)
         elif not releases:
-            self.setCheckState(QtCore.Qt.Unchecked)
+            self.setCheckState(QtCore.Qt.CheckState.Unchecked)
         else:
-            self.setCheckState(QtCore.Qt.PartiallyChecked)
+            self.setCheckState(QtCore.Qt.CheckState.PartiallyChecked)
 
     def nextCheckState(self):
         ids = self.menu.ids
@@ -161,10 +160,10 @@ class CollectionCheckBox(QtWidgets.QCheckBox):
         diff = ids - self.collection.releases
         if diff:
             self.collection.add_releases(diff, self.updateText)
-            self.setCheckState(QtCore.Qt.Checked)
+            self.setCheckState(QtCore.Qt.CheckState.Checked)
         else:
             self.collection.remove_releases(ids & self.collection.releases, self.updateText)
-            self.setCheckState(QtCore.Qt.Unchecked)
+            self.setCheckState(QtCore.Qt.CheckState.Unchecked)
 
     def updateText(self):
         self.setText(self.label())
