@@ -2,9 +2,9 @@
 #
 # Picard, the next-generation MusicBrainz tagger
 #
-# Copyright (C) 2013, 2018, 2020-2021 Laurent Monin
+# Copyright (C) 2013, 2018, 2020-2022 Laurent Monin
 # Copyright (C) 2016-2017 Sambhav Kothari
-# Copyright (C) 2019, 2021 Philipp Wolfer
+# Copyright (C) 2019, 2021-2022 Philipp Wolfer
 # Copyright (C) 2021 Gabriel Ferreira
 #
 # This program is free software; you can redistribute it and/or
@@ -56,7 +56,7 @@ class InfoStatus(QtWidgets.QWidget, Ui_InfoStatus):
         self.label2.setPixmap(self.icon_file.pixmap(size))
         self.label3.setPixmap(self.icon_cd.pixmap(size))
         self.label4.setPixmap(self.icon_file_pending.pixmap(size))
-        self.label5.setPixmap(self.icon_download.pixmap(size, QtGui.QIcon.Disabled))
+        self.label5.setPixmap(self.icon_download.pixmap(size, QtGui.QIcon.Mode.Disabled))
         self._init_tooltips()
 
     def _create_icons(self):
@@ -165,8 +165,8 @@ class InfoStatus(QtWidgets.QWidget, Ui_InfoStatus):
 
     def set_pending_requests(self, num):
         if num <= 0:
-            enabled = QtGui.QIcon.Disabled
+            enabled = QtGui.QIcon.Mode.Disabled
         else:
-            enabled = QtGui.QIcon.Normal
+            enabled = QtGui.QIcon.Mode.Normal
         self.label5.setPixmap(self.icon_download.pixmap(self._size, enabled))
         self.val5.setText(str(num))
