@@ -5,8 +5,9 @@
 # Copyright (C) 2007 Lukáš Lalinský
 # Copyright (C) 2009 Carlin Mangar
 # Copyright (C) 2017 Sambhav Kothari
-# Copyright (C) 2018-2019 Philipp Wolfer
-# Copyright (C) 2018-2020 Laurent Monin
+# Copyright (C) 2018-2022 Philipp Wolfer
+# Copyright (C) 2018-2023 Laurent Monin
+# Copyright (C) 2021 Tche333
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -41,31 +42,30 @@ from PyQt5 import (
     QtCore,
     QtNetwork,
 )
-from PyQt5.QtCore import (
-    QUrl,
-    QUrlQuery,
-)
+from PyQt5.QtCore import QUrl
 from PyQt5.QtNetwork import QNetworkRequest
 
 from picard import (
     PICARD_APP_NAME,
     PICARD_ORG_NAME,
     PICARD_VERSION_STR,
-    config,
     log,
 )
+from picard.config import get_config
 from picard.const import (
-    CACHE_DIR,
     CACHE_SIZE_IN_BYTES,
+    appdirs,
 )
 from picard.oauth import OAuthManager
 from picard.util import (
     build_qurl,
     bytes2human,
+    encoded_queryargs,
     parse_json,
 )
 from picard.util.xml import parse_xml
 from picard.webservice import ratecontrol
+from picard.webservice.utils import port_from_qurl
 
 
 COUNT_REQUESTS_DELAY_MS = 250
@@ -76,9 +76,7 @@ USER_AGENT_STRING = '%s-%s/%s (%s;%s-%s)' % (PICARD_ORG_NAME, PICARD_APP_NAME,
                                              platform.platform(),
                                              platform.python_implementation(),
                                              platform.python_version())
-CLIENT_STRING = bytes(QUrl.toPercentEncoding('%s %s-%s' % (PICARD_ORG_NAME,
-                                                           PICARD_APP_NAME,
-                                                           PICARD_VERSION_STR))).decode()
+CLIENT_STRING = '%s %s-%s' % (PICARD_ORG_NAME, PICARD_APP_NAME, PICARD_VERSION_STR)
 
 
 DEFAULT_RESPONSE_PARSER_TYPE = "json"
@@ -95,73 +93,103 @@ class UnknownResponseParserError(Exception):
 
 class WSRequest(QNetworkRequest):
     """Represents a single HTTP request."""
+    _access_token = None
+    _high_prio_no_cache = True
+    _mblogin = None
+    _retries = 0
 
-    def __init__(self, method, host, port, path, handler, parse_response_type=None, data=None,
-                 mblogin=False, cacheloadcontrol=None, refresh=False,
-                 queryargs=None, priority=False, important=False,
-                 request_mimetype=None):
+    response_mimetype = None
+    response_parser = None
+
+    def __init__(
+        self,
+        *,
+        method=None,
+        handler=None,
+        parse_response_type=None,
+        data=None,
+        mblogin=False,
+        cacheloadcontrol=None,
+        refresh=False,
+        priority=False,
+        important=False,
+        request_mimetype=None,
+        url=None,
+        queryargs=None,
+        unencoded_queryargs=None,
+    ):
         """
         Args:
             method: HTTP method.  One of ``GET``, ``POST``, ``PUT``, or ``DELETE``.
-            host: Hostname.
-            port: TCP port number (80 or 443).
-            path: Path component.
             handler: Callback which takes a 3-tuple of `(str:document,
-            QNetworkReply:reply, QNetworkReply.Error:error)`.
+                QNetworkReply:reply, QNetworkReply.Error:error)`.
             parse_response_type: Specifies that request either sends or accepts
-            data as ``application/{{response_mimetype}}``.
+                data as ``application/{{response_mimetype}}``.
             data: Data to include with ``PUT`` or ``POST`` requests.
             mblogin: Hints that this request should be tied to a MusicBrainz
             account, requiring that we obtain an OAuth token first.
-            cacheloadcontrol: See `QNetworkRequest.CacheLoadControlAttribute`.
+            cacheloadcontrol: See `QNetworkRequest.Attribute.CacheLoadControlAttribute`.
             refresh: Indicates a user-specified resource refresh, such as when
-            the user wishes to reload a release.  Marks the request as high priority
-            and disables caching.
-            queryargs: `dict` of query arguments.
-            retries: Current retry attempt number.
+                the user wishes to reload a release.  Marks the request as high priority
+                and disables caching.
             priority: Indicates that this is a high priority request.
             important: Indicates that this is an important request.
             request_mimetype: Set the Content-Type header.
+            url: URL passed as a string or as a QUrl to use for this request
+            queryargs: Encoded query arguments, a dictionary mapping field names to values
+            unencoded_queryargs: Unencoded query arguments, a dictionary mapping field names to values
         """
-        url = build_qurl(host, port, path=path, queryargs=queryargs)
+        # mandatory parameters
+        self.method = method
+        if self.method not in {'GET', 'PUT', 'DELETE', 'POST'}:
+            raise AssertionError('invalid method')
+
+        self.handler = handler
+        if self.handler is None:
+            raise AssertionError('handler undefined')
+
+        if url is None:
+            raise AssertionError('URL undefined')
+
+        if not isinstance(url, QUrl):
+            url = QUrl(url)
+
+        if queryargs is not None or unencoded_queryargs is not None:
+            if queryargs is None:
+                queryargs = {}
+            if unencoded_queryargs:
+                queryargs.update(encoded_queryargs(unencoded_queryargs))
+            query = QtCore.QUrlQuery(url)
+            for k, v in queryargs.items():
+                query.addQueryItem(k, str(v))
+            url.setQuery(query)
+
         super().__init__(url)
 
-        # These two are codependent (see _update_authorization_header) and must
-        # be initialized explicitly.
-        self._access_token = None
-        self._mblogin = None
-
-        self._retries = 0
-
-        self.method = method
-        self.host = host
-        self.port = port
-        self.path = path
-        self.handler = handler
+        # optional parameters
         self.parse_response_type = parse_response_type
-        self.response_parser = None
-        self.response_mimetype = None
         self.request_mimetype = request_mimetype
         self.data = data
         self.mblogin = mblogin
         self.cacheloadcontrol = cacheloadcontrol
         self.refresh = refresh
-        self.queryargs = queryargs
         self.priority = priority
         self.important = important
-        self._high_prio_no_cache = False
 
-        self.access_token = None
-        self._init_headers()
+        # set headers and attributes
+        self.access_token = None  # call _update_authorization_header
 
-    def _init_headers(self):
-        self.setHeader(QNetworkRequest.UserAgentHeader, USER_AGENT_STRING)
+        if self.method == 'GET':
+            self._high_prio_no_cache = self.refresh
+            self.setAttribute(QNetworkRequest.Attribute.HttpPipeliningAllowedAttribute, True)
+
+        self.setHeader(QNetworkRequest.KnownHeaders.UserAgentHeader, USER_AGENT_STRING)
 
         if self.mblogin or self._high_prio_no_cache:
-            self.setPriority(QNetworkRequest.HighPriority)
-            self.setAttribute(QNetworkRequest.CacheLoadControlAttribute, QNetworkRequest.AlwaysNetwork)
+            self.setPriority(QNetworkRequest.Priority.HighPriority)
+            self.setAttribute(QNetworkRequest.Attribute.CacheLoadControlAttribute, QNetworkRequest.CacheLoadControl.AlwaysNetwork)
         elif self.cacheloadcontrol is not None:
-            self.setAttribute(QNetworkRequest.CacheLoadControlAttribute, self.cacheloadcontrol)
+            self.setAttribute(QNetworkRequest.Attribute.CacheLoadControlAttribute, self.cacheloadcontrol)
 
         if self.parse_response_type:
             try:
@@ -175,13 +203,27 @@ class WSRequest(QNetworkRequest):
         if self.data:
             if not self.request_mimetype:
                 self.request_mimetype = self.response_mimetype or "application/x-www-form-urlencoded"
-            self.setHeader(QNetworkRequest.ContentTypeHeader, self.request_mimetype)
+            self.setHeader(QNetworkRequest.KnownHeaders.ContentTypeHeader, self.request_mimetype)
+
+    @property
+    def has_auth(self):
+        return self.mblogin and self.access_token
 
     def _update_authorization_header(self):
-        authorization = b""
-        if self.mblogin and self.access_token:
-            authorization = ("Bearer %s" % self.access_token).encode('utf-8')
-        self.setRawHeader(b"Authorization", authorization)
+        auth = "Bearer " + self.access_token if self.has_auth else ""
+        self.setRawHeader(b"Authorization", auth.encode('utf-8'))
+
+    @property
+    def host(self):
+        return self.url().host()
+
+    @property
+    def port(self):
+        return port_from_qurl(self.url())
+
+    @property
+    def path(self):
+        return self.url().path()
 
     @property
     def access_token(self):
@@ -217,46 +259,62 @@ class WSRequest(QNetworkRequest):
         return self._retries
 
 
-class WSGetRequest(WSRequest):
+class RequestTask(namedtuple('RequestTask', 'hostkey func priority')):
 
-    def __init__(self, *args, **kwargs):
-        super().__init__("GET", *args, **kwargs)
-
-    def _init_headers(self):
-        self._high_prio_no_cache = self.refresh
-        super()._init_headers()
-        self.setAttribute(QNetworkRequest.HttpPipeliningAllowedAttribute,
-                          True)
+    @staticmethod
+    def from_request(request, func):
+        # priority is a boolean
+        return RequestTask(request.get_host_key(), func, int(request.priority))
 
 
-class WSPutRequest(WSRequest):
+class RequestPriorityQueue:
 
-    def __init__(self, *args, **kwargs):
-        super().__init__("PUT", *args, **kwargs)
+    def __init__(self, ratecontrol):
+        self._queues = defaultdict(lambda: defaultdict(deque))
+        self._ratecontrol = ratecontrol
+        self._count = 0
 
-    def _init_headers(self):
-        self._high_prio_no_cache = True
-        super()._init_headers()
+    def count(self):
+        return self._count
 
+    def add_task(self, task, important=False):
+        (hostkey, func, prio) = task
+        queue = self._queues[prio][hostkey]
+        if important:
+            queue.appendleft(func)
+        else:
+            queue.append(func)
+        self._count += 1
+        return RequestTask(hostkey, func, prio)
 
-class WSDeleteRequest(WSRequest):
+    def remove_task(self, task):
+        hostkey, func, prio = task
+        try:
+            self._queues[prio][hostkey].remove(func)
+            self._count -= 1
+        except Exception as e:
+            log.debug(e)
 
-    def __init__(self, *args, **kwargs):
-        super().__init__("DELETE", *args, **kwargs)
-
-    def _init_headers(self):
-        self._high_prio_no_cache = True
-        super()._init_headers()
-
-
-class WSPostRequest(WSRequest):
-
-    def __init__(self, *args, **kwargs):
-        super().__init__("POST", *args, **kwargs)
-
-    def _init_headers(self):
-        self._high_prio_no_cache = True
-        super()._init_headers()
+    def run_ready_tasks(self):
+        delay = sys.maxsize
+        for prio in sorted(self._queues.keys(), reverse=True):
+            prio_queue = self._queues[prio]
+            if not prio_queue:
+                del self._queues[prio]
+                continue
+            for hostkey in sorted(prio_queue.keys(),
+                                  key=self._ratecontrol.current_delay):
+                queue = self._queues[prio][hostkey]
+                if not queue:
+                    del self._queues[prio][hostkey]
+                    continue
+                wait, d = self._ratecontrol.get_delay_to_next_request(hostkey)
+                if not wait:
+                    queue.popleft()()
+                    self._count -= 1
+                if d < delay:
+                    delay = d
+        return delay
 
 
 class WebService(QtCore.QObject):
@@ -266,9 +324,13 @@ class WebService(QtCore.QObject):
     def __init__(self, parent=None):
         super().__init__(parent)
         self.manager = QtNetwork.QNetworkAccessManager()
+        self._network_accessible_changed(self.manager.networkAccessible())
+        self.manager.networkAccessibleChanged.connect(self._network_accessible_changed)
         self.oauth_manager = OAuthManager(self)
         self.set_cache()
         self.setup_proxy()
+        config = get_config()
+        self.set_transfer_timeout(config.setting['network_transfer_timeout_seconds'])
         self.manager.finished.connect(self._process_reply)
         self._request_methods = {
             "GET": self.manager.get,
@@ -279,11 +341,32 @@ class WebService(QtCore.QObject):
         self._init_queues()
         self._init_timers()
 
+    @staticmethod
+    def http_response_code(reply):
+        response_code = reply.attribute(QNetworkRequest.Attribute.HttpStatusCodeAttribute)
+        return int(response_code) if response_code else 0
+
+    @staticmethod
+    def http_response_phrase(reply):
+        return reply.attribute(QNetworkRequest.Attribute.HttpReasonPhraseAttribute)
+
+    @staticmethod
+    def display_url(url):
+        return url.toDisplayString(QUrl.UrlFormattingOption.RemoveUserInfo | QUrl.ComponentFormattingOption.EncodeSpaces)
+
+    def _network_accessible_changed(self, accessible):
+        # Qt's network accessibility check sometimes fails, e.g. with VPNs on Windows.
+        # If the accessibility is reported to be not accessible, set it to
+        # unknown instead. Let's just try any request and handle the error.
+        # See https://tickets.metabrainz.org/browse/PICARD-1791
+        if accessible == QtNetwork.QNetworkAccessManager.NetworkAccessibility.NotAccessible:
+            self.manager.setNetworkAccessible(QtNetwork.QNetworkAccessManager.NetworkAccessibility.UnknownAccessibility)
+        log.debug("Network accessible requested: %s, actual: %s", accessible, self.manager.networkAccessible())
+
     def _init_queues(self):
         self._active_requests = {}
-        self._queues = defaultdict(lambda: defaultdict(deque))
+        self._queue = RequestPriorityQueue(ratecontrol)
         self.num_pending_web_requests = 0
-        self._last_num_pending_web_requests = -1
 
     def _init_timers(self):
         self._timer_run_next_task = QtCore.QTimer(self)
@@ -297,7 +380,7 @@ class WebService(QtCore.QObject):
         if cache_size_in_bytes is None:
             cache_size_in_bytes = CACHE_SIZE_IN_BYTES
         cache = QtNetwork.QNetworkDiskCache()
-        cache.setCacheDirectory(os.path.join(CACHE_DIR, 'network'))
+        cache.setCacheDirectory(os.path.join(appdirs.cache_folder(), 'network'))
         cache.setMaximumCacheSize(cache_size_in_bytes)
         self.manager.setCache(cache)
         log.debug("NetworkDiskCache dir: %r current size: %s max size: %s",
@@ -307,14 +390,28 @@ class WebService(QtCore.QObject):
 
     def setup_proxy(self):
         proxy = QtNetwork.QNetworkProxy()
+        config = get_config()
         if config.setting["use_proxy"]:
-            proxy.setType(QtNetwork.QNetworkProxy.HttpProxy)
+            if config.setting["proxy_type"] == 'socks':
+                proxy.setType(QtNetwork.QNetworkProxy.ProxyType.Socks5Proxy)
+            else:
+                proxy.setType(QtNetwork.QNetworkProxy.ProxyType.HttpProxy)
             proxy.setHostName(config.setting["proxy_server_host"])
             proxy.setPort(config.setting["proxy_server_port"])
-            proxy.setUser(config.setting["proxy_username"])
-            proxy.setPassword(config.setting["proxy_password"])
+            if config.setting["proxy_username"]:
+                proxy.setUser(config.setting["proxy_username"])
+            if config.setting["proxy_password"]:
+                proxy.setPassword(config.setting["proxy_password"])
         self.manager.setProxy(proxy)
 
+    def set_transfer_timeout(self, timeout):
+        timeout_ms = timeout * 1000
+        if hasattr(self.manager, 'setTransferTimeout'):  # Available since Qt 5.15
+            self.manager.setTransferTimeout(timeout_ms)
+            self._transfer_timeout = 0
+        else:  # Use fallback implementation
+            self._transfer_timeout = timeout_ms
+
     def _send_request(self, request, access_token=None):
         hostkey = request.get_host_key()
         ratecontrol.increment_requests(hostkey)
@@ -323,8 +420,32 @@ class WebService(QtCore.QObject):
         send = self._request_methods[request.method]
         data = request.data
         reply = send(request, data.encode('utf-8')) if data is not None else send(request)
+        self._start_transfer_timeout(reply)
         self._active_requests[reply] = request
 
+    def _start_transfer_timeout(self, reply):
+        if not self._transfer_timeout:
+            return
+        # Fallback implementation of a transfer timeout for Qt < 5.15.
+        # Aborts a request if no data gets transferred for TRANSFER_TIMEOUT milliseconds.
+        timer = QtCore.QTimer(self)
+        timer.setSingleShot(True)
+        timer.setTimerType(QtCore.Qt.TimerType.PreciseTimer)
+        timer.timeout.connect(partial(self._timeout_request, reply))
+        reply.finished.connect(timer.stop)
+        reset_callback = partial(self._reset_transfer_timeout, timer)
+        reply.uploadProgress.connect(reset_callback)
+        reply.downloadProgress.connect(reset_callback)
+        timer.start(self._transfer_timeout)
+
+    def _reset_transfer_timeout(self, timer, bytesTransferred, bytesTotal):
+        timer.start(self._transfer_timeout)
+
+    @staticmethod
+    def _timeout_request(reply):
+        if reply.isRunning():
+            reply.abort()
+
     def _start_request(self, request):
         if request.mblogin and request.path != "/oauth2/token":
             self.oauth_manager.get_access_token(partial(self._send_request, request))
@@ -339,42 +460,38 @@ class WebService(QtCore.QObject):
             from QT 4.7
         """
         return leftUrl.port(80) == rightUrl.port(80) and \
-            leftUrl.toString(QUrl.RemovePort) == rightUrl.toString(QUrl.RemovePort)
-
-    @staticmethod
-    def url_port(url):
-        if url.scheme() == 'https':
-            return url.port(443)
-        return url.port(80)
+            leftUrl.toString(QUrl.UrlFormattingOption.RemovePort) == rightUrl.toString(QUrl.UrlFormattingOption.RemovePort)
 
     def _handle_redirect(self, reply, request, redirect):
-        url = request.url()
         error = int(reply.error())
         # merge with base url (to cover the possibility of the URL being relative)
-        redirect = url.resolved(redirect)
-        if not WebService.urls_equivalent(redirect, reply.request().url()):
-            log.debug("Redirect to %s requested", redirect.toString(QUrl.RemoveUserInfo))
-            redirect_host = redirect.host()
-            redirect_port = self.url_port(redirect)
-            redirect_query = dict(QUrlQuery(redirect).queryItems(QUrl.FullyEncoded))
-            redirect_path = redirect.path()
+        redirect_url = request.url().resolved(redirect)
+        reply_url = reply.request().url()
+        display_redirect_url = self.display_url(redirect_url)
+        display_reply_url = self.display_url(reply_url)
+        if not WebService.urls_equivalent(redirect_url, reply_url):
+            log.debug("Redirect to %s requested", display_redirect_url)
 
-            original_host = url.host()
-            original_port = self.url_port(url)
-            original_host_key = (original_host, original_port)
-            redirect_host_key = (redirect_host, redirect_port)
-            ratecontrol.copy_minimal_delay(original_host_key, redirect_host_key)
+            redirect_request = WSRequest(
+                method='GET',
+                url=redirect_url,
+                handler=request.handler,
+                parse_response_type=request.parse_response_type,
+                priority=True,
+                important=True,
+                mblogin=request.mblogin,
+                cacheloadcontrol=request.attribute(QNetworkRequest.Attribute.CacheLoadControlAttribute),
+                refresh=request.refresh,
+            )
 
-            self.get(redirect_host,
-                     redirect_port,
-                     redirect_path,
-                     request.handler, request.parse_response_type, priority=True, important=True,
-                     refresh=request.refresh, queryargs=redirect_query, mblogin=request.mblogin,
-                     cacheloadcontrol=request.attribute(QNetworkRequest.CacheLoadControlAttribute))
+            ratecontrol.copy_minimal_delay(
+                request.get_host_key(),
+                redirect_request.get_host_key(),
+            )
+
+            self.add_request(redirect_request)
         else:
-            log.error("Redirect loop: %s",
-                      reply.request().url().toString(QUrl.RemoveUserInfo)
-                      )
+            log.error("Redirect loop: %s", display_reply_url)
             request.handler(reply.readAll(), reply, error)
 
     def _handle_reply(self, reply, request):
@@ -387,16 +504,15 @@ class WebService(QtCore.QObject):
 
         error = int(reply.error())
         handler = request.handler
+        response_code = self.http_response_code(reply)
+        display_reply_url = self.display_url(reply.request().url())
         if error:
-            code = reply.attribute(QNetworkRequest.HttpStatusCodeAttribute)
-            code = int(code) if code else 0
             errstr = reply.errorString()
-            url = reply.request().url().toString(QUrl.RemoveUserInfo)
-            log.error("Network request error for %s: %s (QT code %d, HTTP code %d)",
-                      url, errstr, error, code)
+            log.error("Network request error for %s -> %s (QT code %d, HTTP code %d)",
+                      display_reply_url, errstr, error, response_code)
             if (not request.max_retries_reached()
-                and (code == 503
-                     or code == 429
+                and (response_code == 503
+                     or response_code == 429
                      # Sometimes QT returns a http status code of 200 even when there
                      # is a service unavailable error. But it returns a QT error code
                      # of 403 when this happens
@@ -404,22 +520,22 @@ class WebService(QtCore.QObject):
                      )):
                 slow_down = True
                 retries = request.mark_for_retry()
-                log.debug("Retrying %s (#%d)", url, retries)
+                log.debug("Retrying %s (#%d)", display_reply_url, retries)
                 self.add_request(request)
 
             elif handler is not None:
                 handler(reply.readAll(), reply, error)
 
-            slow_down = (slow_down or code >= 500)
+            slow_down = (slow_down or response_code >= 500)
 
         else:
-            redirect = reply.attribute(QNetworkRequest.RedirectionTargetAttribute)
-            fromCache = reply.attribute(QNetworkRequest.SourceIsFromCacheAttribute)
-            cached = ' (CACHED)' if fromCache else ''
-            log.debug("Received reply for %s: HTTP %d (%s) %s",
-                      reply.request().url().toString(QUrl.RemoveUserInfo),
-                      reply.attribute(QNetworkRequest.HttpStatusCodeAttribute),
-                      reply.attribute(QNetworkRequest.HttpReasonPhraseAttribute),
+            redirect = reply.attribute(QNetworkRequest.Attribute.RedirectionTargetAttribute)
+            from_cache = reply.attribute(QNetworkRequest.Attribute.SourceIsFromCacheAttribute)
+            cached = ' (CACHED)' if from_cache else ''
+            log.debug("Received reply for %s -> HTTP %d (%s) %s",
+                      display_reply_url,
+                      response_code,
+                      self.http_response_phrase(reply),
                       cached
                       )
             if handler is not None:
@@ -429,9 +545,9 @@ class WebService(QtCore.QObject):
                 elif request.response_parser:
                     try:
                         document = request.response_parser(reply)
+                        log.debug("Response received: %s", document)
                     except Exception as e:
-                        url = reply.request().url().toString(QUrl.RemoveUserInfo)
-                        log.error("Unable to parse the response for %s: %s", url, e)
+                        log.error("Unable to parse the response for %s -> %s", display_reply_url, e)
                         document = reply.readAll()
                         error = e
                     finally:
@@ -445,7 +561,8 @@ class WebService(QtCore.QObject):
         try:
             request = self._active_requests.pop(reply)
         except KeyError:
-            log.error("Request not found for %s", reply.request().url().toString(QUrl.RemoveUserInfo))
+            display_reply_url = self.display_url(reply.request().url())
+            log.error("Request not found for %s", display_reply_url)
             return
         try:
             self._handle_reply(reply, request)
@@ -456,40 +573,107 @@ class WebService(QtCore.QObject):
     def get(self, host, port, path, handler, parse_response_type=DEFAULT_RESPONSE_PARSER_TYPE,
             priority=False, important=False, mblogin=False, cacheloadcontrol=None, refresh=False,
             queryargs=None):
-        request = WSGetRequest(host, port, path, handler, parse_response_type=parse_response_type,
-                               mblogin=mblogin, cacheloadcontrol=cacheloadcontrol, refresh=refresh,
-                               queryargs=queryargs, priority=priority, important=important)
+        log.warning("This method is deprecated, use WebService.get_url() instead")
+        request = WSRequest(
+            method='GET',
+            url=build_qurl(host, port, path=path, queryargs=queryargs),
+            handler=handler,
+            parse_response_type=parse_response_type,
+            priority=priority,
+            important=important,
+            mblogin=mblogin,
+            cacheloadcontrol=cacheloadcontrol,
+            refresh=refresh,
+        )
         return self.add_request(request)
 
     def post(self, host, port, path, data, handler, parse_response_type=DEFAULT_RESPONSE_PARSER_TYPE,
              priority=False, important=False, mblogin=True, queryargs=None, request_mimetype=None):
-        request = WSPostRequest(host, port, path, handler, parse_response_type=parse_response_type,
-                                data=data, mblogin=mblogin, queryargs=queryargs,
-                                priority=priority, important=important,
-                                request_mimetype=request_mimetype)
+        log.warning("This method is deprecated, use WebService.post_url() instead")
+        request = WSRequest(
+            method='POST',
+            url=build_qurl(host, port, path=path, queryargs=queryargs),
+            handler=handler,
+            parse_response_type=parse_response_type,
+            priority=priority,
+            important=important,
+            mblogin=mblogin,
+            data=data,
+            request_mimetype=request_mimetype,
+        )
         log.debug("POST-DATA %r", data)
         return self.add_request(request)
 
     def put(self, host, port, path, data, handler, priority=True, important=False, mblogin=True,
             queryargs=None, request_mimetype=None):
-        request = WSPutRequest(host, port, path, handler, data=data, mblogin=mblogin,
-                               queryargs=queryargs, priority=priority,
-                               important=important, request_mimetype=request_mimetype)
+        log.warning("This method is deprecated, use WebService.put_url() instead")
+        request = WSRequest(
+            method='PUT',
+            url=build_qurl(host, port, path=path, queryargs=queryargs),
+            handler=handler,
+            priority=priority,
+            important=important,
+            mblogin=mblogin,
+            data=data,
+            request_mimetype=request_mimetype,
+        )
         return self.add_request(request)
 
     def delete(self, host, port, path, handler, priority=True, important=False, mblogin=True,
                queryargs=None):
-        request = WSDeleteRequest(host, port, path, handler, mblogin=mblogin,
-                                  queryargs=queryargs, priority=priority, important=important)
+        log.warning("This method is deprecated, use WebService.delete_url() instead")
+        request = WSRequest(
+            method='DELETE',
+            url=build_qurl(host, port, path=path, queryargs=queryargs),
+            handler=handler,
+            priority=priority,
+            important=important,
+            mblogin=mblogin,
+        )
         return self.add_request(request)
 
     def download(self, host, port, path, handler, priority=False,
                  important=False, cacheloadcontrol=None, refresh=False,
                  queryargs=None):
-        return self.get(host, port, path, handler, parse_response_type=None,
-                        priority=priority, important=important,
-                        cacheloadcontrol=cacheloadcontrol, refresh=refresh,
-                        queryargs=queryargs)
+        log.warning("This method is deprecated, use WebService.download_url() instead")
+        request = WSRequest(
+            method='GET',
+            url=build_qurl(host, port, path=path, queryargs=queryargs),
+            handler=handler,
+            priority=priority,
+            important=important,
+            cacheloadcontrol=cacheloadcontrol,
+            refresh=refresh,
+        )
+        return self.add_request(request)
+
+    def get_url(self, **kwargs):
+        kwargs['method'] = 'GET'
+        kwargs['parse_response_type'] = kwargs.get('parse_response_type', DEFAULT_RESPONSE_PARSER_TYPE)
+        return self.add_request(WSRequest(**kwargs))
+
+    def post_url(self, **kwargs):
+        kwargs['method'] = 'POST'
+        kwargs['parse_response_type'] = kwargs.get('parse_response_type', DEFAULT_RESPONSE_PARSER_TYPE)
+        kwargs['mblogin'] = kwargs.get('mblogin', True)
+        log.debug("POST-DATA %r", kwargs['data'])
+        return self.add_request(WSRequest(**kwargs))
+
+    def put_url(self, **kwargs):
+        kwargs['method'] = 'PUT'
+        kwargs['priority'] = kwargs.get('priority', True)
+        kwargs['mblogin'] = kwargs.get('mblogin', True)
+        return self.add_request(WSRequest(**kwargs))
+
+    def delete_url(self, **kwargs):
+        kwargs['method'] = 'DELETE'
+        kwargs['priority'] = kwargs.get('priority', True)
+        kwargs['mblogin'] = kwargs.get('mblogin', True)
+        return self.add_request(WSRequest(**kwargs))
+
+    def download_url(self, **kwargs):
+        kwargs['method'] = 'GET'
+        return self.add_request(WSRequest(**kwargs))
 
     def stop(self):
         for reply in list(self._active_requests):
@@ -497,45 +681,21 @@ class WebService(QtCore.QObject):
         self._init_queues()
 
     def _count_pending_requests(self):
-        count = len(self._active_requests)
-        for prio_queue in self._queues.values():
-            for queue in prio_queue.values():
-                count += len(queue)
-        self.num_pending_web_requests = count
-        if count != self._last_num_pending_web_requests:
-            self._last_num_pending_web_requests = count
+        count = len(self._active_requests) + self._queue.count()
+        if count != self.num_pending_web_requests:
+            self.num_pending_web_requests = count
             self.tagger.tagger_stats_changed.emit()
         if count:
             self._timer_count_pending_requests.start(COUNT_REQUESTS_DELAY_MS)
 
     def _run_next_task(self):
-        delay = sys.maxsize
-        for prio in sorted(self._queues.keys(), reverse=True):
-            prio_queue = self._queues[prio]
-            if not prio_queue:
-                del(self._queues[prio])
-                continue
-            for hostkey in sorted(prio_queue.keys(),
-                                  key=ratecontrol.current_delay):
-                queue = self._queues[prio][hostkey]
-                if not queue:
-                    del(self._queues[prio][hostkey])
-                    continue
-                wait, d = ratecontrol.get_delay_to_next_request(hostkey)
-                if not wait:
-                    queue.popleft()()
-                if d < delay:
-                    delay = d
+        delay = self._queue.run_ready_tasks()
         if delay < sys.maxsize:
             self._timer_run_next_task.start(delay)
 
     def add_task(self, func, request):
-        hostkey = request.get_host_key()
-        prio = int(request.priority)  # priority is a boolean
-        if request.important:
-            self._queues[prio][hostkey].appendleft(func)
-        else:
-            self._queues[prio][hostkey].append(func)
+        task = RequestTask.from_request(request, func)
+        self._queue.add_task(task, request.important)
 
         if not self._timer_run_next_task.isActive():
             self._timer_run_next_task.start(0)
@@ -543,19 +703,15 @@ class WebService(QtCore.QObject):
         if not self._timer_count_pending_requests.isActive():
             self._timer_count_pending_requests.start(0)
 
-        return (hostkey, func, prio)
+        return task
 
     def add_request(self, request):
         return self.add_task(partial(self._start_request, request), request)
 
     def remove_task(self, task):
-        hostkey, func, prio = task
-        try:
-            self._queues[prio][hostkey].remove(func)
-            if not self._timer_count_pending_requests.isActive():
-                self._timer_count_pending_requests.start(0)
-        except Exception as e:
-            log.debug(e)
+        self._queue.remove_task(task)
+        if not self._timer_count_pending_requests.isActive():
+            self._timer_count_pending_requests.start(0)
 
     @classmethod
     def add_parser(cls, response_type, mimetype, parser):
